#   Copyright 2012-2013 OpenStack Foundation
#   Copyright 2015 Dean Troyer
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Command-line interface to the OpenStack APIs"""

import sys
import warnings

from osc_lib.api import auth
from osc_lib.command import commandmanager
from osc_lib import shell

import openstackclient
from openstackclient.common import clientmanager

DEFAULT_DOMAIN = 'default'
# list of modules that were originally out-of-tree and are now in
# core OSC
IGNORED_MODULES = (
    'neutron_taas.taas_client.osc',
    'neutronclient.osc.v2.taas',
)


class OpenStackShell(shell.OpenStackShell):
    client_manager: clientmanager.ClientManager

    def __init__(self):
        command_manager = commandmanager.CommandManager(
            'openstack.cli', ignored_modules=IGNORED_MODULES
        )

        super().__init__(
            description=__doc__.strip(),
            version=openstackclient.__version__,
            command_manager=command_manager,
            deferred_help=True,
        )

        self.api_version = {}

        # Assume TLS host certificate verification is enabled
        self.verify = True

        # ignore warnings from openstacksdk since our users can't do anything
        # about them
        warnings.filterwarnings('ignore', module='openstack')

    def build_option_parser(self, description, version, argparse_kwargs=None):
        parser = super().build_option_parser(
            description, version, argparse_kwargs
        )
        parser = clientmanager.build_plugin_option_parser(parser)
        parser = auth.build_auth_plugins_option_parser(parser)
        return parser

    def _final_defaults(self):
        super()._final_defaults()

        # Set the default plugin to admin_token if endpoint and token are given
        if self.options.endpoint and self.options.token:
            # Use token authentication
            self._auth_type = 'admin_token'
        else:
            self._auth_type = 'password'

    def _load_plugins(self):
        """Load plugins via stevedore."""
        # Loop through extensions to get API versions
        for mod in clientmanager.PLUGIN_MODULES:
            default_version = getattr(mod, 'DEFAULT_API_VERSION', None)
            # Only replace the first instance of "os", some service names will
            # have "os" in their name, like: "antiddos"
            option = mod.API_VERSION_OPTION.replace('os_', '', 1)
            version_opt = str(self.cloud.config.get(option, default_version))
            if version_opt:
                api = mod.API_NAME
                self.api_version[api] = version_opt

                # Add a plugin interface to let the module validate the version
                # requested by the user
                skip_old_check = False
                mod_check_api_version = getattr(mod, 'check_api_version', None)
                if mod_check_api_version:
                    # this throws an exception if invalid
                    skip_old_check = mod_check_api_version(version_opt)

                # NOTE(stephenfin): API_VERSIONS has traditionally been a
                # dictionary but the values are only used internally and are
                # ignored for the modules using SDK. So we now support tuples
                # instead.
                mod_versions = getattr(mod, 'API_VERSIONS', None)
                if mod_versions is not None and not isinstance(
                    mod_versions, dict | tuple
                ):
                    raise TypeError(
                        f'Plugin {mod} has incompatible API_VERSIONS. '
                        f'Expected: tuple, dict. Got: {type(mod_versions)}. '
                        f'Please report this to your package maintainer.'
                    )

                if mod_versions and not skip_old_check:
                    if version_opt not in mod_versions:
                        sorted_versions = sorted(
                            list(mod.API_VERSIONS),
                            key=lambda s: list(map(int, s.split('.'))),
                        )
                        self.log.warning(
                            "%(name)s API version %(version)s is not in "
                            "supported versions: %(supported)s",
                            {
                                'name': api,
                                'version': version_opt,
                                'supported': ', '.join(sorted_versions),
                            },
                        )

                # Command groups deal only with major versions
                version = '.v' + version_opt.replace('.', '_').split('_')[0]
                cmd_group = 'openstack.' + api.replace('-', '_') + version
                self.command_manager.add_command_group(cmd_group)
                self.log.debug(
                    '%(name)s API version %(version)s, cmd group %(group)s',
                    {'name': api, 'version': version_opt, 'group': cmd_group},
                )

    def _load_commands(self):
        """Load commands via cliff/stevedore

        osc-lib has no opinion on what commands should be loaded
        """
        # Commands that span multiple APIs
        self.command_manager.add_command_group('openstack.common')

        # This is the naive extension implementation referred to in
        # blueprint 'client-extensions'
        # Extension modules can register their commands in an
        # 'openstack.extension' entry point group:
        # entry_points={
        #     'openstack.extension': [
        #         'list_repo=qaz.github.repo:ListRepo',
        #         'show_repo=qaz.github.repo:ShowRepo',
        #     ],
        # }
        self.command_manager.add_command_group('openstack.extension')

    def initialize_app(self, argv):
        super().initialize_app(argv)

        # Re-create the client_manager with our subclass
        self.client_manager = clientmanager.ClientManager(
            cli_options=self.cloud,
            api_version=self.api_version,
            pw_func=shell.prompt_for_password,
        )


def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    return OpenStackShell().run(argv)


if __name__ == "__main__":
    sys.exit(main())



#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest import mock

from openstack.block_storage.v2 import service as _service
from openstack.test import fakes as sdk_fakes
from osc_lib import exceptions

from openstackclient.tests.unit.volume.v2 import fakes as volume_fakes
from openstackclient.volume.v2 import service


class TestServiceList(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        self.service = sdk_fakes.generate_fake_resource(_service.Service)
        self.volume_sdk_client.services.return_value = [self.service]

        self.cmd = service.ListService(self.app, None)

    def test_service_list(self):
        arglist = [
            '--host',
            self.service.host,
            '--service',
            self.service.binary,
        ]
        verifylist = [
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        expected_columns = (
            'Binary',
            'Host',
            'Zone',
            'Status',
            'State',
            'Updated At',
        )
        datalist = (
            (
                self.service.binary,
                self.service.host,
                self.service.availability_zone,
                self.service.status,
                self.service.state,
                self.service.updated_at,
            ),
        )
        self.assertEqual(expected_columns, columns)
        self.assertEqual(datalist, tuple(data))
        self.volume_sdk_client.services.assert_called_with(
            host=self.service.host,
            binary=self.service.binary,
        )

    def test_service_list_with_long_option(self):
        arglist = [
            '--host',
            self.service.host,
            '--service',
            self.service.binary,
            '--long',
        ]
        verifylist = [
            ('host', self.service.host),
            ('service', self.service.binary),
            ('long', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        expected_columns = (
            'Binary',
            'Host',
            'Zone',
            'Status',
            'State',
            'Updated At',
            'Disabled Reason',
        )
        datalist = (
            (
                self.service.binary,
                self.service.host,
                self.service.availability_zone,
                self.service.status,
                self.service.state,
                self.service.updated_at,
                self.service.disabled_reason,
            ),
        )
        self.assertEqual(expected_columns, columns)
        self.assertEqual(datalist, tuple(data))
        self.volume_sdk_client.services.assert_called_with(
            host=self.service.host,
            binary=self.service.binary,
        )


class TestServiceSet(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        self.service = sdk_fakes.generate_fake_resource(_service.Service)
        self.service.enable = mock.Mock(autospec=True)
        self.service.disable = mock.Mock(autospec=True)
        self.volume_sdk_client.find_service.return_value = self.service

        self.cmd = service.SetService(self.app, None)

    def test_service_set_nothing(self):
        arglist = [
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.service.enable.assert_not_called()
        self.service.disable.assert_not_called()
        self.assertIsNone(result)

    def test_service_set_enable(self):
        arglist = [
            '--enable',
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('enable', True),
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.service.enable.assert_called_with(self.volume_sdk_client)
        self.service.disable.assert_not_called()
        self.assertIsNone(result)

    def test_service_set_disable(self):
        arglist = [
            '--disable',
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('disable', True),
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.service.enable.assert_not_called()
        self.service.disable.assert_called_with(
            self.volume_sdk_client, reason=None
        )
        self.assertIsNone(result)

    def test_service_set_disable_with_reason(self):
        reason = 'earthquake'
        arglist = [
            '--disable',
            '--disable-reason',
            reason,
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('disable', True),
            ('disable_reason', reason),
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.service.enable.assert_not_called()
        self.service.disable.assert_called_with(
            self.volume_sdk_client, reason=reason
        )
        self.assertIsNone(result)

    def test_service_set_only_with_disable_reason(self):
        reason = 'earthquake'
        arglist = [
            '--disable-reason',
            reason,
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('disable_reason', reason),
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        try:
            self.cmd.take_action(parsed_args)
            self.fail("CommandError should be raised.")
        except exceptions.CommandError as e:
            self.assertEqual(
                "Cannot specify option --disable-reason without "
                "--disable specified.",
                str(e),
            )

    def test_service_set_enable_with_disable_reason(self):
        reason = 'earthquake'
        arglist = [
            '--enable',
            '--disable-reason',
            reason,
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('enable', True),
            ('disable_reason', reason),
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        try:
            self.cmd.take_action(parsed_args)
            self.fail("CommandError should be raised.")
        except exceptions.CommandError as e:
            self.assertEqual(
                "Cannot specify option --disable-reason without "
                "--disable specified.",
                str(e),
            )



#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Volume v3 Backup action implementations"""

import copy
import functools
import logging

from cliff import columns as cliff_columns
from openstack import utils as sdk_utils
from osc_lib.cli import parseractions
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.common import pagination
from openstackclient.i18n import _

LOG = logging.getLogger(__name__)


class VolumeIdColumn(cliff_columns.FormattableColumn[str]):
    """Formattable column for volume ID column.

    Unlike the parent FormattableColumn class, the initializer of the
    class takes volume_cache as the second argument.
    osc_lib.utils.get_item_properties instantiate cliff FormattableColumn
    object with a single parameter "column value", so you need to pass
    a partially initialized class like
    ``functools.partial(VolumeIdColumn, volume_cache)``.
    """

    def __init__(self, value, volume_cache=None):
        super().__init__(value)
        self._volume_cache = volume_cache or {}

    def human_readable(self):
        """Return a volume name if available

        :rtype: either the volume ID or name
        """
        volume_id = self._value
        volume = volume_id
        if volume_id in self._volume_cache.keys():
            volume = self._volume_cache[volume_id].name
        return volume


class CreateVolumeBackup(command.ShowOne):
    _description = _("Create new volume backup")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "volume",
            metavar="<volume>",
            help=_("Volume to backup (name or ID)"),
        )
        parser.add_argument(
            "--name", metavar="<name>", help=_("Name of the backup")
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            help=_("Description of the backup"),
        )
        parser.add_argument(
            "--container",
            metavar="<container>",
            help=_("Optional backup container name"),
        )
        parser.add_argument(
            "--snapshot",
            metavar="<snapshot>",
            help=_("Snapshot to backup (name or ID)"),
        )
        parser.add_argument(
            '--force',
            action='store_true',
            default=False,
            help=_("Allow to back up an in-use volume"),
        )
        parser.add_argument(
            '--incremental',
            action='store_true',
            default=False,
            help=_("Perform an incremental backup"),
        )
        parser.add_argument(
            '--no-incremental',
            action='store_false',
            help=_("Do not perform an incremental backup"),
        )
        parser.add_argument(
            '--property',
            metavar='<key=value>',
            action=parseractions.KeyValueAction,
            dest='properties',
            help=_(
                'Set a property on this backup '
                '(repeat option to remove multiple values) '
                '(supported by --os-volume-api-version 3.43 or above)'
            ),
        )
        parser.add_argument(
            '--availability-zone',
            metavar='<zone-name>',
            help=_(
                'AZ where the backup should be stored; by default it will be '
                'the same as the source '
                '(supported by --os-volume-api-version 3.51 or above)'
            ),
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.sdk_connection.volume

        volume_id = volume_client.find_volume(
            parsed_args.volume,
            ignore_missing=False,
        ).id

        kwargs = {}

        if parsed_args.snapshot:
            kwargs['snapshot_id'] = volume_client.find_snapshot(
                parsed_args.snapshot,
                ignore_missing=False,
            ).id

        if parsed_args.properties:
            if not sdk_utils.supports_microversion(volume_client, '3.43'):
                msg = _(
                    '--os-volume-api-version 3.43 or greater is required to '
                    'support the --property option'
                )
                raise exceptions.CommandError(msg)

            kwargs['metadata'] = parsed_args.properties

        if parsed_args.availability_zone:
            if not sdk_utils.supports_microversion(volume_client, '3.51'):
                msg = _(
                    '--os-volume-api-version 3.51 or greater is required to '
                    'support the --availability-zone option'
                )
                raise exceptions.CommandError(msg)

            kwargs['availability_zone'] = parsed_args.availability_zone

        columns: tuple[str, ...] = (
            "id",
            "name",
            "volume_id",
        )
        backup = volume_client.create_backup(
            volume_id=volume_id,
            container=parsed_args.container,
            name=parsed_args.name,
            description=parsed_args.description,
            force=parsed_args.force,
            is_incremental=parsed_args.incremental,
            **kwargs,
        )
        data = utils.get_dict_properties(backup, columns)
        return (columns, data)


class DeleteVolumeBackup(command.Command):
    _description = _("Delete volume backup(s)")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "backups",
            metavar="<backup>",
            nargs="+",
            help=_("Backup(s) to delete (name or ID)"),
        )
        parser.add_argument(
            '--force',
            action='store_true',
            default=False,
            help=_("Allow delete in state other than error or available"),
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.sdk_connection.volume
        result = 0

        for backup in parsed_args.backups:
            try:
                backup_id = volume_client.find_backup(
                    backup, ignore_missing=False
                ).id
                volume_client.delete_backup(
                    backup_id,
                    ignore_missing=False,
                    force=parsed_args.force,
                )
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete backup with "
                        "name or ID '%(backup)s': %(e)s"
                    )
                    % {'backup': backup, 'e': e}
                )

        if result > 0:
            total = len(parsed_args.backups)
            msg = _("%(result)s of %(total)s backups failed to delete.") % {
                'result': result,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ListVolumeBackup(command.Lister):
    _description = _("List volume backups")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_('Filter results by project (name or ID) (admin only)'),
        )
        parser.add_argument(
            "--long",
            action="store_true",
            default=False,
            help=_("List additional fields in output"),
        )
        parser.add_argument(
            "--name",
            metavar="<name>",
            help=_("Filters results by the backup name"),
        )
        parser.add_argument(
            "--status",
            metavar="<status>",
            choices=[
                'creating',
                'available',
                'deleting',
                'error',
                'restoring',
                'error_restoring',
            ],
            help=_(
                "Filters results by the backup status, one of: "
                "creating, available, deleting, error, restoring or "
                "error_restoring"
            ),
        )
        parser.add_argument(
            "--volume",
            metavar="<volume>",
            help=_(
                "Filters results by the volume which they backup (name or ID)"
            ),
        )
        pagination.add_marker_pagination_option_to_parser(parser)
        parser.add_argument(
            '--all-projects',
            action='store_true',
            default=False,
            help=_('Include all projects (admin only)'),
        )
        # TODO(stephenfin): Add once we have an equivalent command for
        # 'cinder list-filters'
        # parser.add_argument(
        #     '--filter',
        #     metavar='<key=value>',
        #     action=parseractions.KeyValueAction,
        #     dest='filters',
        #     help=_(
        #         "Filter key and value pairs. Use 'foo' to "
        #         "check enabled filters from server. Use 'key~=value' for "
        #         "inexact filtering if the key supports "
        #         "(supported by --os-volume-api-version 3.33 or above)"
        #     ),
        # )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.sdk_connection.volume
        identity_client = self.app.client_manager.sdk_connection.identity

        columns: tuple[str, ...] = (
            'id',
            'name',
            'description',
            'status',
            'size',
            'is_incremental',
            'created_at',
        )
        column_headers: tuple[str, ...] = (
            'ID',
            'Name',
            'Description',
            'Status',
            'Size',
            'Incremental',
            'Created At',
        )
        if parsed_args.long:
            columns += ('availability_zone', 'volume_id', 'container')
            column_headers += ('Availability Zone', 'Volume', 'Container')

        # Cache the volume list
        volume_cache = {}
        try:
            for s in volume_client.volumes():
                volume_cache[s.id] = s
        except Exception:  # noqa: S110
            # Just forget it if there's any trouble
            pass

        _VolumeIdColumn = functools.partial(
            VolumeIdColumn, volume_cache=volume_cache
        )

        all_tenants = parsed_args.all_projects
        project_id = None
        if parsed_args.project:
            all_tenants = True
            project_id = identity_client.find_project(
                parsed_args.project, ignore_missing=False
            ).id

        filter_volume_id = None
        if parsed_args.volume:
            try:
                filter_volume_id = volume_client.find_volume(
                    parsed_args.volume,
                    ignore_missing=False,
                ).id
            except exceptions.CommandError:
                # Volume with that ID does not exist, but search for backups
                # for that volume nevertheless
                LOG.debug(
                    "No volume with ID %s existing, continuing to "
                    "search for backups for that volume ID",
                    parsed_args.volume,
                )
                filter_volume_id = parsed_args.volume

        marker_backup_id = None
        if parsed_args.marker:
            marker_backup_id = volume_client.find_backup(
                parsed_args.marker,
                ignore_missing=False,
            ).id

        data = volume_client.backups(
            name=parsed_args.name,
            status=parsed_args.status,
            volume_id=filter_volume_id,
            all_tenants=all_tenants,
            marker=marker_backup_id,
            limit=parsed_args.limit,
            project_id=project_id,
        )

        return (
            column_headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters={'volume_id': _VolumeIdColumn},
                )
                for s in data
            ),
        )


class RestoreVolumeBackup(command.ShowOne):
    _description = _("Restore volume backup")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "backup",
            metavar="<backup>",
            help=_("Backup to restore (name or ID)"),
        )
        parser.add_argument(
            "volume",
            metavar="<volume>",
            nargs="?",
            help=_(
                "Volume to restore to "
                "(name or ID for existing volume, name only for new volume) "
                "(default to None)"
            ),
        )
        parser.add_argument(
            "--force",
            action="store_true",
            help=_(
                "Restore the backup to an existing volume (default to False)"
            ),
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.sdk_connection.volume

        columns: tuple[str, ...] = (
            'id',
            'volume_id',
            'volume_name',
        )

        backup = volume_client.find_backup(
            parsed_args.backup,
            ignore_missing=False,
        )

        volume_name = None
        volume_id = None
        try:
            volume_id = volume_client.find_volume(
                parsed_args.volume,
                ignore_missing=False,
            ).id
        except Exception:
            volume_name = parsed_args.volume
        else:
            # If we didn't fail, the volume must already exist. We only allow
            # this to work if the user forced things
            if not parsed_args.force:
                msg = _(
                    "Volume '%s' already exists; if you want to restore the "
                    "backup to it you need to specify the '--force' option"
                )
                raise exceptions.CommandError(msg % parsed_args.volume)

        restore = volume_client.restore_backup(
            backup.id,
            volume_id=volume_id,
            name=volume_name,
        )

        data = utils.get_dict_properties(restore, columns)
        return (columns, data)


class SetVolumeBackup(command.Command):
    _description = _("Set volume backup properties")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "backup",
            metavar="<backup>",
            help=_("Backup to modify (name or ID)"),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_(
                'New backup name '
                '(supported by --os-volume-api-version 3.9 or above)'
            ),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_(
                'New backup description '
                '(supported by --os-volume-api-version 3.9 or above)'
            ),
        )
        parser.add_argument(
            '--state',
            metavar='<state>',
            choices=['available', 'error'],
            help=_(
                'New backup state ("available" or "error") (admin only) '
                '(This option simply changes the state of the backup '
                'in the database with no regard to actual status; '
                'exercise caution when using)'
            ),
        )
        parser.add_argument(
            '--no-property',
            action='store_true',
            help=_(
                'Remove all properties from this backup '
                '(specify both --no-property and --property to remove the '
                'current properties before setting new properties)'
            ),
        )
        parser.add_argument(
            '--property',
            metavar='<key=value>',
            action=parseractions.KeyValueAction,
            dest='properties',
            default={},
            help=_(
                'Set a property on this backup '
                '(repeat option to set multiple values) '
                '(supported by --os-volume-api-version 3.43 or above)'
            ),
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.sdk_connection.volume

        backup = volume_client.find_backup(
            parsed_args.backup,
            ignore_missing=False,
        )

        result = 0
        if parsed_args.state:
            try:
                volume_client.reset_backup_status(
                    backup, status=parsed_args.state
                )
            except Exception as e:
                LOG.error(_("Failed to set backup state: %s"), e)
                result += 1

        kwargs = {}

        if parsed_args.name:
            if not sdk_utils.supports_microversion(volume_client, '3.9'):
                msg = _(
                    '--os-volume-api-version 3.9 or greater is required to '
                    'support the --name option'
                )
                raise exceptions.CommandError(msg)

            kwargs['name'] = parsed_args.name

        if parsed_args.description:
            if not sdk_utils.supports_microversion(volume_client, '3.9'):
                msg = _(
                    '--os-volume-api-version 3.9 or greater is required to '
                    'support the --description option'
                )
                raise exceptions.CommandError(msg)

            kwargs['description'] = parsed_args.description

        if parsed_args.no_property:
            if not sdk_utils.supports_microversion(volume_client, '3.43'):
                msg = _(
                    '--os-volume-api-version 3.43 or greater is required to '
                    'support the --no-property option'
                )
                raise exceptions.CommandError(msg)

        if parsed_args.properties:
            if not sdk_utils.supports_microversion(volume_client, '3.43'):
                msg = _(
                    '--os-volume-api-version 3.43 or greater is required to '
                    'support the --property option'
                )
                raise exceptions.CommandError(msg)

        if sdk_utils.supports_microversion(volume_client, '3.43'):
            metadata = copy.deepcopy(backup.metadata)

            if parsed_args.no_property:
                metadata = {}

            metadata.update(parsed_args.properties)
            kwargs['metadata'] = metadata

        if kwargs:
            try:
                volume_client.update_backup(backup, **kwargs)
            except Exception as e:
                LOG.error("Failed to update backup: %s", e)
                result += 1

        if result > 0:
            msg = _("One or more of the set operations failed")
            raise exceptions.CommandError(msg)


class UnsetVolumeBackup(command.Command):
    """Unset volume backup properties.

    This command requires ``--os-volume-api-version`` 3.43 or greater.
    """

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'backup',
            metavar='<backup>',
            help=_('Backup to modify (name or ID)'),
        )
        parser.add_argument(
            '--property',
            metavar='<key>',
            action='append',
            dest='properties',
            help=_(
                'Property to remove from this backup '
                '(repeat option to unset multiple values) '
            ),
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.sdk_connection.volume

        if not sdk_utils.supports_microversion(volume_client, '3.43'):
            msg = _(
                '--os-volume-api-version 3.43 or greater is required to '
                'support the --property option'
            )
            raise exceptions.CommandError(msg)

        backup = volume_client.find_backup(
            parsed_args.backup, ignore_missing=False
        )
        metadata = copy.deepcopy(backup.metadata)

        for key in parsed_args.properties:
            if key not in metadata:
                # ignore invalid properties but continue
                LOG.warning(
                    "'%s' is not a valid property for backup '%s'",
                    key,
                    parsed_args.backup,
                )
                continue

            del metadata[key]

        volume_client.delete_backup_metadata(backup, keys=list(metadata))


class ShowVolumeBackup(command.ShowOne):
    _description = _("Display volume backup details")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "backup",
            metavar="<backup>",
            help=_("Backup to display (name or ID)"),
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.sdk_connection.volume
        backup = volume_client.find_backup(
            parsed_args.backup, ignore_missing=False
        )
        columns: tuple[str, ...] = (
            "availability_zone",
            "container",
            "created_at",
            "data_timestamp",
            "description",
            "encryption_key_id",
            "fail_reason",
            "has_dependent_backups",
            "id",
            "is_incremental",
            "metadata",
            "name",
            "object_count",
            "project_id",
            "size",
            "snapshot_id",
            "status",
            "updated_at",
            "user_id",
            "volume_id",
        )
        data = utils.get_dict_properties(backup, columns)
        return (columns, data)



#   Copyright 2019 SUSE LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Identity v3 Access Rule action implementations"""

import logging

from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common


LOG = logging.getLogger(__name__)


class DeleteAccessRule(command.Command):
    _description = _("Delete access rule(s)")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'access_rule',
            metavar='<access-rule>',
            nargs="+",
            help=_('Access rule ID(s) to delete'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.sdk_connection.identity
        conn = self.app.client_manager.sdk_connection
        auth = conn.config.get_auth()
        if auth is None:
            # this will never happen
            raise exceptions.CommandError('invalid authentication info')
        user_id = auth.get_user_id(conn.identity)

        errors = 0
        for ac in parsed_args.access_rule:
            try:
                access_rule = identity_client.get_access_rule(user_id, ac)
                identity_client.delete_access_rule(user_id, access_rule.id)
            except Exception as e:
                errors += 1
                LOG.error(
                    _("Failed to delete access rule with ID '%(ac)s': %(e)s"),
                    {'ac': ac, 'e': e},
                )

        if errors > 0:
            total = len(parsed_args.access_rule)
            msg = _(
                "%(errors)s of %(total)s access rules failed to delete."
            ) % {'errors': errors, 'total': total}
            raise exceptions.CommandError(msg)


class ListAccessRule(command.Lister):
    _description = _("List access rules")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--user',
            metavar='<user>',
            help=_('User whose access rules to list (name or ID)'),
        )
        common.add_user_domain_option_to_parser(parser)
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.sdk_connection.identity
        if parsed_args.user:
            user_id = common.find_user(
                identity_client, parsed_args.user, parsed_args.user_domain
            ).id
        else:
            conn = self.app.client_manager.sdk_connection
            auth = conn.config.get_auth()
            if auth is None:
                # this will never happen
                raise exceptions.CommandError('invalid authentication info')
            user_id = auth.get_user_id(conn.identity)

        columns = ('ID', 'Service', 'Method', 'Path')
        data = identity_client.access_rules(user=user_id)
        return (
            columns,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters={},
                )
                for s in data
            ),
        )


class ShowAccessRule(command.ShowOne):
    _description = _("Display access rule details")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'access_rule',
            metavar='<access-rule>',
            help=_('Access rule ID to display'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.sdk_connection.identity
        conn = self.app.client_manager.sdk_connection
        auth = conn.config.get_auth()
        if auth is None:
            # this will never happen
            raise exceptions.CommandError('invalid authentication info')
        user_id = auth.get_user_id(conn.identity)

        access_rule = identity_client.get_access_rule(
            user_id, parsed_args.access_rule
        )

        columns = ('ID', 'Method', 'Path', 'Service')
        return (
            columns,
            (
                utils.get_item_properties(
                    access_rule,
                    columns,
                    formatters={},
                )
            ),
        )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Registered limits action implementations."""

import logging

from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common as common_utils

LOG = logging.getLogger(__name__)


class CreateRegisteredLimit(command.ShowOne):
    _description = _("Create a registered limit")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('Description of the registered limit'),
        )
        parser.add_argument(
            '--region',
            metavar='<region>',
            help=_('Region for the registered limit to affect'),
        )
        parser.add_argument(
            '--service',
            metavar='<service>',
            required=True,
            help=_(
                'Service responsible for the resource to limit (required) '
                '(name or ID)'
            ),
        )
        parser.add_argument(
            '--default-limit',
            type=int,
            metavar='<default-limit>',
            required=True,
            help=_('The default limit for the resources to assume (required)'),
        )
        parser.add_argument(
            'resource_name',
            metavar='<resource-name>',
            help=_('The name of the resource to limit'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity

        service = utils.find_resource(
            identity_client.services, parsed_args.service
        )
        region = None
        if parsed_args.region:
            if 'None' not in parsed_args.region:
                # NOTE (vishakha): Due to bug #1799153 and for any another
                # related case where GET resource API does not support the
                # filter by name, osc_lib.utils.find_resource() method cannot
                # be used because that method try to fall back to list all the
                # resource if requested resource cannot be get via name. Which
                # ends up with NoUniqueMatch error.
                # So osc_lib.utils.find_resource() function cannot be used for
                # 'regions', using common_utils.get_resource() instead.
                region = common_utils.get_resource(
                    identity_client.regions, parsed_args.region
                )
            else:
                self.log.warning(
                    _(
                        "Passing 'None' to indicate no region is deprecated. "
                        "Instead, don't pass --region."
                    )
                )

        registered_limit = identity_client.registered_limits.create(
            service,
            parsed_args.resource_name,
            parsed_args.default_limit,
            description=parsed_args.description,
            region=region,
        )

        registered_limit._info.pop('links', None)
        return zip(*sorted(registered_limit._info.items()))


class DeleteRegisteredLimit(command.Command):
    _description = _("Delete a registered limit")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'registered_limits',
            metavar='<registered-limits>',
            nargs="+",
            help=_('Registered limit(s) to delete (ID)'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity

        errors = 0
        for registered_limit_id in parsed_args.registered_limits:
            try:
                identity_client.registered_limits.delete(registered_limit_id)
            except Exception as e:
                errors += 1
                from pprint import pprint

                pprint(type(e))
                LOG.error(
                    _(
                        "Failed to delete registered limit with ID "
                        "'%(id)s': %(e)s"
                    ),
                    {'id': registered_limit_id, 'e': e},
                )

        if errors > 0:
            total = len(parsed_args.registered_limits)
            msg = _(
                "%(errors)s of %(total)s registered limits failed to delete."
            ) % {'errors': errors, 'total': total}
            raise exceptions.CommandError(msg)


class ListRegisteredLimit(command.Lister):
    _description = _("List registered limits")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--service',
            metavar='<service>',
            help=_(
                'Service responsible for the resource to limit (name or ID)'
            ),
        )
        parser.add_argument(
            '--resource-name',
            metavar='<resource-name>',
            dest='resource_name',
            help=_('The name of the resource to limit'),
        )
        parser.add_argument(
            '--region',
            metavar='<region>',
            help=_('Region for the limit to affect.'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity

        service = None
        if parsed_args.service:
            service = common_utils.find_service(
                identity_client, parsed_args.service
            )
        region = None
        if parsed_args.region:
            if 'None' not in parsed_args.region:
                # NOTE (vishakha): Due to bug #1799153 and for any another
                # related case where GET resource API does not support the
                # filter by name, osc_lib.utils.find_resource() method cannot
                # be used because that method try to fall back to list all the
                # resource if requested resource cannot be get via name. Which
                # ends up with NoUniqueMatch error.
                # So osc_lib.utils.find_resource() function cannot be used for
                # 'regions', using common_utils.get_resource() instead.
                region = common_utils.get_resource(
                    identity_client.regions, parsed_args.region
                )
            else:
                self.log.warning(
                    _(
                        "Passing 'None' to indicate no region is deprecated. "
                        "Instead, don't pass --region."
                    )
                )

        registered_limits = identity_client.registered_limits.list(
            service=service,
            resource_name=parsed_args.resource_name,
            region=region,
        )

        columns = (
            'ID',
            'Service ID',
            'Resource Name',
            'Default Limit',
            'Description',
            'Region ID',
        )
        return (
            columns,
            (utils.get_item_properties(s, columns) for s in registered_limits),
        )


class SetRegisteredLimit(command.ShowOne):
    _description = _("Update information about a registered limit")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'registered_limit_id',
            metavar='<registered-limit-id>',
            help=_('Registered limit to update (ID)'),
        )
        parser.add_argument(
            '--service',
            metavar='<service>',
            help=_(
                'Service to be updated responsible for the resource to limit '
                '(name or ID). Either --service, --resource-name or --region '
                'must be different than existing value otherwise it will be '
                'duplicate entry'
            ),
        )
        parser.add_argument(
            '--resource-name',
            metavar='<resource-name>',
            help=_(
                'Resource to be updated responsible for the resource to '
                'limit. Either --service, --resource-name or --region must '
                'be different than existing value otherwise it will be '
                'duplicate entry'
            ),
        )
        parser.add_argument(
            '--default-limit',
            metavar='<default-limit>',
            type=int,
            help=_('The default limit for the resources to assume'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('Description to update of the registered limit'),
        )
        parser.add_argument(
            '--region',
            metavar='<region>',
            help=_(
                'Region for the registered limit to affect. Either '
                '--service, --resource-name or --region must be '
                'different than existing value otherwise it will be '
                'duplicate entry'
            ),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity

        service = None
        if parsed_args.service:
            service = common_utils.find_service(
                identity_client, parsed_args.service
            )

        region = None
        if parsed_args.region:
            if 'None' not in parsed_args.region:
                # NOTE (vishakha): Due to bug #1799153 and for any another
                # related case where GET resource API does not support the
                # filter by name, osc_lib.utils.find_resource() method cannot
                # be used because that method try to fall back to list all the
                # resource if requested resource cannot be get via name. Which
                # ends up with NoUniqueMatch error.
                # So osc_lib.utils.find_resource() function cannot be used for
                # 'regions', using common_utils.get_resource() instead.
                region = common_utils.get_resource(
                    identity_client.regions, parsed_args.region
                )
            else:
                self.log.warning(
                    _("Passing 'None' to indicate no region is deprecated.")
                )

        registered_limit = identity_client.registered_limits.update(
            parsed_args.registered_limit_id,
            service=service,
            resource_name=parsed_args.resource_name,
            default_limit=parsed_args.default_limit,
            description=parsed_args.description,
            region=region,
        )

        registered_limit._info.pop('links', None)
        return zip(*sorted(registered_limit._info.items()))


class ShowRegisteredLimit(command.ShowOne):
    _description = _("Display registered limit details")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'registered_limit_id',
            metavar='<registered-limit-id>',
            help=_('Registered limit to display (ID)'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity
        registered_limit = identity_client.registered_limits.get(
            parsed_args.registered_limit_id
        )
        registered_limit._info.pop('links', None)
        return zip(*sorted(registered_limit._info.items()))



#   Copyright 2013 Nebula Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest import mock

from openstack import exceptions as sdk_exc
from openstack.identity.v3 import domain as _domain
from openstack.identity.v3 import group as _group
from openstack.identity.v3 import project as _project
from openstack.identity.v3 import role as _role
from openstack.identity.v3 import system as _system
from openstack.identity.v3 import user as _user
from openstack.test import fakes as sdk_fakes
from osc_lib import exceptions

from openstackclient.identity.v3 import role
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes


class TestRoleInherited(identity_fakes.TestIdentityv3):
    def _is_inheritance_testcase(self):
        return True


class TestRoleAdd(identity_fakes.TestIdentityv3):
    def _is_inheritance_testcase(self):
        return False

    user = sdk_fakes.generate_fake_resource(_user.User)
    group = sdk_fakes.generate_fake_resource(_group.Group)
    domain = sdk_fakes.generate_fake_resource(_domain.Domain)
    project = sdk_fakes.generate_fake_resource(_project.Project)

    def setUp(self):
        super().setUp()

        self.identity_sdk_client.find_user.return_value = self.user
        self.identity_sdk_client.find_group.return_value = self.group
        self.identity_sdk_client.find_domain.return_value = self.domain
        self.identity_sdk_client.find_project.return_value = self.project

        self.role = sdk_fakes.generate_fake_resource(
            resource_type=_role.Role,
            domain_id=None,
            description=None,
        )
        self.identity_sdk_client.find_role.return_value = self.role
        self.role_with_domain = sdk_fakes.generate_fake_resource(
            resource_type=_role.Role,
            domain_id=self.domain.id,
            description=None,
        )

        self.identity_sdk_client.assign_domain_role_to_user.return_value = (
            self.role
        )
        self.identity_sdk_client.assign_domain_role_to_group.return_value = (
            self.role
        )
        self.identity_sdk_client.assign_project_role_to_user.return_value = (
            self.role
        )
        self.identity_sdk_client.assign_project_role_to_group.return_value = (
            self.role
        )
        self.identity_sdk_client.assign_system_role_to_user.return_value = (
            self.role
        )
        self.identity_sdk_client.assign_system_role_to_group.return_value = (
            self.role
        )

        # Get the command object to test
        self.cmd = role.AddRole(self.app, None)

    @mock.patch.object(role.LOG, 'warning')
    def test_role_add_user_system(self, mock_warning):
        arglist = [
            '--user',
            self.user.name,
            '--system',
            'all',
            self.role.name,
        ]
        if self._is_inheritance_testcase():
            arglist.append('--inherited')
        verifylist = [
            ('user', self.user.name),
            ('group', None),
            ('system', 'all'),
            ('domain', None),
            ('project', None),
            ('role', self.role.name),
            ('inherited', self._is_inheritance_testcase()),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'system': 'all',
            'user': self.user.id,
            'role': self.role.id,
        }
        self.identity_sdk_client.assign_system_role_to_user.assert_called_with(
            **kwargs
        )
        self.assertIsNone(result)

        if self._is_inheritance_testcase():
            mock_warning.assert_called_with(
                "'--inherited' was given, which is not supported when adding a system role; this will be an error in a future release"
            )

    def test_role_add_user_domain(self):
        arglist = [
            '--user',
            self.user.name,
            '--domain',
            self.domain.name,
            self.role.name,
        ]
        if self._is_inheritance_testcase():
            arglist.append('--inherited')
        verifylist = [
            ('user', self.user.name),
            ('group', None),
            ('domain', self.domain.name),
            ('project', None),
            ('role', self.role.name),
            ('inherited', self._is_inheritance_testcase()),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'domain': self.domain.id,
            'user': self.user.id,
            'role': self.role.id,
            'inherited': self._is_inheritance_testcase(),
        }
        self.identity_sdk_client.assign_domain_role_to_user.assert_called_with(
            **kwargs
        )
        self.assertIsNone(result)

    def test_role_add_user_project(self):
        arglist = [
            '--user',
            self.user.name,
            '--project',
            self.project.name,
            self.role.name,
        ]
        if self._is_inheritance_testcase():
            arglist.append('--inherited')
        verifylist = [
            ('user', self.user.name),
            ('group', None),
            ('domain', None),
            ('project', self.project.name),
            ('role', self.role.name),
            ('inherited', self._is_inheritance_testcase()),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'project': self.project.id,
            'user': self.user.id,
            'role': self.role.id,
            'inherited': self._is_inheritance_testcase(),
        }
        self.identity_sdk_client.assign_project_role_to_user.assert_called_with(
            **kwargs
        )
        self.assertIsNone(result)

    @mock.patch.object(role.LOG, 'warning')
    def test_role_add_group_system(self, mock_warning):
        arglist = [
            '--group',
            self.group.name,
            '--system',
            'all',
            self.role.name,
        ]
        if self._is_inheritance_testcase():
            arglist.append('--inherited')
        verifylist = [
            ('user', None),
            ('group', self.group.name),
            ('system', 'all'),
            ('domain', None),
            ('project', None),
            ('role', self.role.name),
            ('inherited', self._is_inheritance_testcase()),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'system': 'all',
            'group': self.group.id,
            'role': self.role.id,
        }
        self.identity_sdk_client.assign_system_role_to_group.assert_called_with(
            **kwargs
        )
        self.assertIsNone(result)

        if self._is_inheritance_testcase():
            mock_warning.assert_called_with(
                "'--inherited' was given, which is not supported when adding a system role; this will be an error in a future release"
            )

    def test_role_add_group_domain(self):
        arglist = [
            '--group',
            self.group.name,
            '--domain',
            self.domain.name,
            self.role.name,
        ]
        if self._is_inheritance_testcase():
            arglist.append('--inherited')
        verifylist = [
            ('user', None),
            ('group', self.group.name),
            ('domain', self.domain.name),
            ('project', None),
            ('role', self.role.name),
            ('inherited', self._is_inheritance_testcase()),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'domain': self.domain.id,
            'group': self.group.id,
            'role': self.role.id,
            'inherited': self._is_inheritance_testcase(),
        }
        self.identity_sdk_client.assign_domain_role_to_group.assert_called_with(
            **kwargs
        )
        self.assertIsNone(result)

    def test_role_add_group_project(self):
        arglist = [
            '--group',
            self.group.name,
            '--project',
            self.project.name,
            self.role.name,
        ]
        if self._is_inheritance_testcase():
            arglist.append('--inherited')
        verifylist = [
            ('user', None),
            ('group', self.group.name),
            ('domain', None),
            ('project', self.project.name),
            ('role', self.role.name),
            ('inherited', self._is_inheritance_testcase()),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'project': self.project.id,
            'group': self.group.id,
            'role': self.role.id,
            'inherited': self._is_inheritance_testcase(),
        }
        self.identity_sdk_client.assign_project_role_to_group.assert_called_with(
            **kwargs
        )
        self.assertIsNone(result)

    def test_role_add_domain_role_on_user_project(self):
        self.identity_sdk_client.find_role.return_value = self.role_with_domain

        arglist = [
            '--user',
            self.user.name,
            '--project',
            self.project.name,
            '--role-domain',
            self.domain.name,
            self.role_with_domain.name,
        ]
        if self._is_inheritance_testcase():
            arglist.append('--inherited')
        verifylist = [
            ('user', self.user.name),
            ('group', None),
            ('domain', None),
            ('project', self.project.name),
            ('role', self.role_with_domain.name),
            ('inherited', self._is_inheritance_testcase()),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'project': self.project.id,
            'user': self.user.id,
            'role': self.role_with_domain.id,
            'inherited': self._is_inheritance_testcase(),
        }
        self.identity_sdk_client.assign_project_role_to_user.assert_called_with(
            **kwargs
        )
        self.assertIsNone(result)

    def test_role_add_with_error(self):
        arglist = [
            self.role.name,
        ]
        verifylist = [
            ('user', None),
            ('group', None),
            ('domain', None),
            ('project', None),
            ('role', self.role.name),
            ('inherited', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestRoleAddInherited(TestRoleAdd, TestRoleInherited):
    pass


class TestRoleCreate(identity_fakes.TestIdentityv3):
    collist = ('id', 'name', 'domain_id', 'description')
    domain = sdk_fakes.generate_fake_resource(_domain.Domain)

    def setUp(self):
        super().setUp()

        self.role = sdk_fakes.generate_fake_resource(
            resource_type=_role.Role,
            domain_id=None,
            description=None,
        )
        self.role_with_domain = sdk_fakes.generate_fake_resource(
            resource_type=_role.Role,
            domain_id=self.domain.id,
            description=None,
        )
        self.role_with_description = sdk_fakes.generate_fake_resource(
            resource_type=_role.Role,
            domain_id=None,
            description='role description',
        )
        self.role_with_immutable_option = sdk_fakes.generate_fake_resource(
            resource_type=_role.Role,
            domain_id=None,
            description=None,
            options={'immutable': True},
        )
        self.identity_sdk_client.find_domain.return_value = self.domain

        # Get the command object to test
        self.cmd = role.CreateRole(self.app, None)

    def test_role_create_no_options(self):
        self.identity_sdk_client.create_role.return_value = self.role

        arglist = [
            self.role.name,
        ]
        verifylist = [
            ('name', self.role.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.role.name,
        }

        self.identity_sdk_client.create_role.assert_called_with(**kwargs)

        self.assertEqual(self.collist, columns)
        datalist = (
            self.role.id,
            self.role.name,
            None,
            None,
        )
        self.assertEqual(datalist, data)

    def test_role_create_with_domain(self):
        self.identity_sdk_client.create_role.return_value = (
            self.role_with_domain
        )

        arglist = [
            '--domain',
            self.domain.name,
            self.role_with_domain.name,
        ]
        verifylist = [
            ('domain', self.domain.name),
            ('name', self.role_with_domain.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'domain_id': self.domain.id,
            'name': self.role_with_domain.name,
        }

        self.identity_sdk_client.create_role.assert_called_with(**kwargs)

        self.assertEqual(self.collist, columns)
        datalist = (
            self.role_with_domain.id,
            self.role_with_domain.name,
            self.domain.id,
            None,
        )
        self.assertEqual(datalist, data)

    def test_role_create_with_description(self):
        self.identity_sdk_client.create_role.return_value = (
            self.role_with_description
        )

        arglist = [
            '--description',
            self.role_with_description.description,
            self.role_with_description.name,
        ]
        verifylist = [
            ('description', self.role_with_description.description),
            ('name', self.role_with_description.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.role_with_description.name,
            'description': self.role_with_description.description,
        }

        self.identity_sdk_client.create_role.assert_called_with(**kwargs)

        self.assertEqual(self.collist, columns)
        datalist = (
            self.role_with_description.id,
            self.role_with_description.name,
            None,
            self.role_with_description.description,
        )
        self.assertEqual(datalist, data)

    def test_role_create_with_immutable_option(self):
        self.identity_sdk_client.create_role.return_value = self.role

        arglist = [
            '--immutable',
            self.role.name,
        ]
        verifylist = [
            ('immutable', True),
            ('name', self.role.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'options': {'immutable': True},
            'name': self.role.name,
        }

        self.identity_sdk_client.create_role.assert_called_with(**kwargs)

        self.assertEqual(self.collist, columns)
        datalist = (
            self.role.id,
            self.role.name,
            None,
            None,
        )
        self.assertEqual(datalist, data)

    def test_role_create_with_no_immutable_option(self):
        self.identity_sdk_client.create_role.return_value = self.role

        arglist = [
            '--no-immutable',
            self.role.name,
        ]
        verifylist = [
            ('immutable', False),
            ('name', self.role.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'options': {'immutable': False},
            'name': self.role.name,
        }

        self.identity_sdk_client.create_role.assert_called_with(**kwargs)

        self.assertEqual(self.collist, columns)
        datalist = (
            self.role.id,
            self.role.name,
            None,
            None,
        )
        self.assertEqual(datalist, data)


class TestRoleDelete(identity_fakes.TestIdentityv3):
    def setUp(self):
        super().setUp()

        # Get the command object to test
        self.cmd = role.DeleteRole(self.app, None)

    def test_role_delete_no_options(self):
        self.role = sdk_fakes.generate_fake_resource(
            resource_type=_role.Role,
            domain_id=None,
            description=None,
        )
        self.identity_sdk_client.find_role.return_value = self.role
        self.identity_sdk_client.delete_role.return_value = None

        arglist = [
            self.role.name,
        ]
        verifylist = [
            ('roles', [self.role.name]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.delete_role.assert_called_with(
            role=self.role.id,
            ignore_missing=False,
        )
        self.assertIsNone(result)

    def test_role_delete_with_domain(self):
        self.domain = sdk_fakes.generate_fake_resource(_domain.Domain)
        self.role_with_domain = sdk_fakes.generate_fake_resource(
            resource_type=_role.Role,
            domain_id=self.domain.id,
            description=None,
        )
        self.identity_sdk_client.find_role.return_value = self.role_with_domain
        self.identity_sdk_client.delete_role.return_value = None

        arglist = [
            '--domain',
            self.domain.name,
            self.role_with_domain.name,
        ]
        verifylist = [
            ('roles', [self.role_with_domain.name]),
            ('domain', self.domain.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.delete_role.assert_called_with(
            role=self.role_with_domain.id,
            ignore_missing=False,
        )
        self.assertIsNone(result)

    def test_delete_multi_roles_with_exception(self):
        self.role = sdk_fakes.generate_fake_resource(
            resource_type=_role.Role,
            domain_id=None,
            description=None,
        )
        self.identity_sdk_client.find_role.side_effect = [
            self.role,
            sdk_exc.ResourceNotFound,
        ]
        arglist = [
            self.role.id,
            'unexist_role',
        ]
        verifylist = [
            ('roles', arglist),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual('1 of 2 roles failed to delete.', str(e))

        self.identity_sdk_client.find_role.assert_has_calls(
            [
                mock.call(
                    name_or_id=self.role.id,
                    ignore_missing=False,
                    domain_id=None,
                ),
                mock.call(
                    name_or_id='unexist_role',
                    ignore_missing=False,
                    domain_id=None,
                ),
            ]
        )

        self.assertEqual(2, self.identity_sdk_client.find_role.call_count)
        self.identity_sdk_client.delete_role.assert_called_once_with(
            role=self.role.id, ignore_missing=False
        )


class TestRoleList(identity_fakes.TestIdentityv3):
    columns = (
        'ID',
        'Name',
    )
    domain = sdk_fakes.generate_fake_resource(_domain.Domain)

    def setUp(self):
        super().setUp()

        self.role = sdk_fakes.generate_fake_resource(
            resource_type=_role.Role,
            domain_id=None,
            description=None,
        )
        self.role_with_domain = sdk_fakes.generate_fake_resource(
            resource_type=_role.Role,
            domain_id=self.domain.id,
            description=None,
        )
        self.identity_sdk_client.roles.return_value = [
            self.role,
            self.role_with_domain,
        ]
        self.identity_sdk_client.find_domain.return_value = self.domain

        self.datalist = (
            (
                self.role.id,
                self.role.name,
            ),
            (
                self.role_with_domain.id,
                self.role_with_domain.name,
            ),
        )

        # Get the command object to test
        self.cmd = role.ListRole(self.app, None)

    def test_role_list_no_options(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.roles.assert_called_with()

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_role_list_domain_role(self):
        self.identity_sdk_client.roles.return_value = [self.role_with_domain]
        arglist = [
            '--domain',
            self.domain.name,
        ]
        verifylist = [
            ('domain', self.domain.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {'domain_id': self.domain.id}
        self.identity_sdk_client.roles.assert_called_with(**kwargs)

        collist = ('ID', 'Name', 'Domain')
        self.assertEqual(collist, columns)
        datalist = (
            (
                self.role_with_domain.id,
                self.role_with_domain.name,
                self.domain.name,
            ),
        )
        self.assertEqual(datalist, tuple(data))


class TestRoleRemove(identity_fakes.TestIdentityv3):
    def _is_inheritance_testcase(self):
        return False

    user = sdk_fakes.generate_fake_resource(_user.User)
    group = sdk_fakes.generate_fake_resource(_group.Group)
    domain = sdk_fakes.generate_fake_resource(_domain.Domain)
    project = sdk_fakes.generate_fake_resource(_project.Project)
    system = sdk_fakes.generate_fake_resource(_system.System)

    def setUp(self):
        super().setUp()

        self.role = sdk_fakes.generate_fake_resource(
            resource_type=_role.Role,
            domain_id=None,
            description=None,
        )
        self.identity_sdk_client.find_role.return_value = self.role
        self.identity_sdk_client.find_user.return_value = self.user
        self.identity_sdk_client.find_group.return_value = self.group
        self.identity_sdk_client.find_domain.return_value = self.domain
        self.identity_sdk_client.find_project.return_value = self.project

        self.identity_sdk_client.unassign_domain_role_from_user.return_value = None
        self.identity_sdk_client.unassign_domain_role_from_group.return_value = None
        self.identity_sdk_client.unassign_project_role_from_user.return_value = None
        self.identity_sdk_client.unassign_project_role_from_group.return_value = None
        self.identity_sdk_client.unassign_system_role_from_user.return_value = None
        self.identity_sdk_client.unassign_system_role_from_group.return_value = None

        # Get the command object to test
        self.cmd = role.RemoveRole(self.app, None)

    def test_role_remove_user_system(self):
        arglist = [
            '--user',
            self.user.name,
            '--system',
            'all',
            self.role.name,
        ]
        if self._is_inheritance_testcase():
            arglist.append('--inherited')
        verifylist = [
            ('user', self.user.name),
            ('group', None),
            ('system', 'all'),
            ('domain', None),
            ('project', None),
            ('role', self.role.name),
            ('inherited', self._is_inheritance_testcase()),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'user': self.user.id,
            'system': 'all',
            'role': self.role.id,
        }
        self.identity_sdk_client.unassign_system_role_from_user.assert_called_with(
            **kwargs
        )
        self.assertIsNone(result)

    def test_role_remove_non_existent_user_system(self):
        # Simulate the user not being in keystone; the client should gracefully
        # handle this exception and send the request to remove the role since
        # keystone supports removing role assignments with non-existent actors
        # (e.g., users or groups).
        self.identity_sdk_client.find_user.side_effect = [
            sdk_exc.ResourceNotFound,
        ]
        arglist = [
            '--user',
            self.user.id,
            '--system',
            'all',
            self.role.name,
        ]
        if self._is_inheritance_testcase():
            arglist.append('--inherited')
        verifylist = [
            ('user', self.user.id),
            ('group', None),
            ('system', 'all'),
            ('domain', None),
            ('project', None),
            ('role', self.role.name),
            ('inherited', self._is_inheritance_testcase()),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'user': self.user.id,
            'system': 'all',
            'role': self.role.id,
        }
        self.identity_sdk_client.unassign_system_role_from_user.assert_called_with(
            **kwargs
        )
        self.assertIsNone(result)

    def test_role_remove_user_domain(self):
        arglist = [
            '--user',
            self.user.name,
            '--domain',
            self.domain.name,
            self.role.name,
        ]
        if self._is_inheritance_testcase():
            arglist.append('--inherited')
        verifylist = [
            ('user', self.user.name),
            ('group', None),
            ('domain', self.domain.name),
            ('project', None),
            ('role', self.role.name),
            ('inherited', self._is_inheritance_testcase()),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'user': self.user.id,
            'domain': self.domain.id,
            'role': self.role.id,
            'inherited': self._is_inheritance_testcase(),
        }
        self.identity_sdk_client.unassign_domain_role_from_user.assert_called_with(
            **kwargs
        )
        self.assertIsNone(result)

    def test_role_remove_non_existent_user_domain(self):
        # Simulate the user not being in keystone, the client the gracefully
        # handle this exception and send the request to remove the role since
        # keystone will validate.
        self.identity_sdk_client.find_user.side_effect = [
            sdk_exc.ResourceNotFound,
        ]
        arglist = [
            '--user',
            self.user.id,
            '--domain',
            self.domain.name,
            self.role.name,
        ]
        if self._is_inheritance_testcase():
            arglist.append('--inherited')
        verifylist = [
            ('user', self.user.id),
            ('group', None),
            ('system', None),
            ('domain', self.domain.name),
            ('project', None),
            ('role', self.role.name),
            ('inherited', self._is_inheritance_testcase()),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'user': self.user.id,
            'domain': self.domain.id,
            'role': self.role.id,
            'inherited': self._is_inheritance_testcase(),
        }
        self.identity_sdk_client.unassign_domain_role_from_user.assert_called_with(
            **kwargs
        )
        self.assertIsNone(result)

    def test_role_remove_user_project(self):
        arglist = [
            '--user',
            self.user.name,
            '--project',
            self.project.name,
            self.role.name,
        ]
        if self._is_inheritance_testcase():
            arglist.append('--inherited')
        verifylist = [
            ('user', self.user.name),
            ('group', None),
            ('domain', None),
            ('project', self.project.name),
            ('role', self.role.name),
            ('inherited', self._is_inheritance_testcase()),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'user': self.user.id,
            'project': self.project.id,
            'role': self.role.id,
            'inherited': self._is_inheritance_testcase(),
        }
        self.identity_sdk_client.unassign_project_role_from_user.assert_called_with(
            **kwargs
        )
        self.assertIsNone(result)

    def test_role_remove_non_existent_user_project(self):
        # Simulate the user not being in keystone, the client the gracefully
        # handle this exception and send the request to remove the role since
        # keystone will validate.
        self.identity_sdk_client.find_user.side_effect = [
            sdk_exc.ResourceNotFound,
        ]

        arglist = [
            '--user',
            self.user.id,
            '--project',
            self.project.name,
            self.role.name,
        ]
        if self._is_inheritance_testcase():
            arglist.append('--inherited')
        verifylist = [
            ('user', self.user.id),
            ('group', None),
            ('system', None),
            ('domain', None),
            ('project', self.project.name),
            ('role', self.role.name),
            ('inherited', self._is_inheritance_testcase()),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'user': self.user.id,
            'project': self.project.id,
            'role': self.role.id,
            'inherited': self._is_inheritance_testcase(),
        }
        self.identity_sdk_client.unassign_project_role_from_user.assert_called_with(
            **kwargs
        )
        self.assertIsNone(result)

    def test_role_remove_group_system(self):
        arglist = [
            '--group',
            self.group.name,
            '--system',
            'all',
            self.role.name,
        ]
        if self._is_inheritance_testcase():
            arglist.append('--inherited')
        verifylist = [
            ('user', None),
            ('group', self.group.name),
            ('system', 'all'),
            ('domain', None),
            ('project', None),
            ('role', self.role.name),
            ('role', self.role.name),
            ('inherited', self._is_inheritance_testcase()),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'group': self.group.id,
            'system': 'all',
            'role': self.role.id,
        }
        self.identity_sdk_client.unassign_system_role_from_group.assert_called_with(
            **kwargs
        )
        self.assertIsNone(result)

    def test_role_remove_non_existent_group_system(self):
        # Simulate the user not being in keystone, the client the gracefully
        # handle this exception and send the request to remove the role since
        # keystone will validate.
        self.identity_sdk_client.find_group.side_effect = [
            sdk_exc.ResourceNotFound,
        ]

        arglist = [
            '--group',
            self.group.id,
            '--system',
            'all',
            self.role.name,
        ]
        if self._is_inheritance_testcase():
            arglist.append('--inherited')
        verifylist = [
            ('user', None),
            ('group', self.group.id),
            ('system', 'all'),
            ('domain', None),
            ('project', None),
            ('role', self.role.name),
            ('inherited', self._is_inheritance_testcase()),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'group': self.group.id,
            'system': 'all',
            'role': self.role.id,
        }
        self.identity_sdk_client.unassign_system_role_from_group.assert_called_with(
            **kwargs
        )
        self.assertIsNone(result)

    def test_role_remove_group_domain(self):
        arglist = [
            '--group',
            self.group.name,
            '--domain',
            self.domain.name,
            self.role.name,
        ]
        if self._is_inheritance_testcase():
            arglist.append('--inherited')
        verifylist = [
            ('user', None),
            ('group', self.group.name),
            ('domain', self.domain.name),
            ('project', None),
            ('role', self.role.name),
            ('role', self.role.name),
            ('inherited', self._is_inheritance_testcase()),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'group': self.group.id,
            'domain': self.domain.id,
            'role': self.role.id,
            'inherited': self._is_inheritance_testcase(),
        }
        self.identity_sdk_client.unassign_domain_role_from_group.assert_called_with(
            **kwargs
        )
        self.assertIsNone(result)

    def test_role_remove_non_existent_group_domain(self):
        # Simulate the user not being in keystone, the client the gracefully
        # handle this exception and send the request to remove the role since
        # keystone will validate.
        self.identity_sdk_client.find_group.side_effect = [
            sdk_exc.ResourceNotFound,
        ]

        arglist = [
            '--group',
            self.group.id,
            '--domain',
            self.domain.name,
            self.role.name,
        ]
        if self._is_inheritance_testcase():
            arglist.append('--inherited')
        verifylist = [
            ('user', None),
            ('group', self.group.id),
            ('system', None),
            ('domain', self.domain.name),
            ('project', None),
            ('role', self.role.name),
            ('inherited', self._is_inheritance_testcase()),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'group': self.group.id,
            'domain': self.domain.id,
            'role': self.role.id,
            'inherited': self._is_inheritance_testcase(),
        }
        self.identity_sdk_client.unassign_domain_role_from_group.assert_called_with(
            **kwargs
        )
        self.assertIsNone(result)

    def test_role_remove_group_project(self):
        arglist = [
            '--group',
            self.group.name,
            '--project',
            self.project.name,
            self.role.name,
        ]
        if self._is_inheritance_testcase():
            arglist.append('--inherited')
        verifylist = [
            ('user', None),
            ('group', self.group.name),
            ('domain', None),
            ('project', self.project.name),
            ('role', self.role.name),
            ('inherited', self._is_inheritance_testcase()),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'group': self.group.id,
            'project': self.project.id,
            'role': self.role.id,
            'inherited': self._is_inheritance_testcase(),
        }
        self.identity_sdk_client.unassign_project_role_from_group.assert_called_with(
            **kwargs
        )
        self.assertIsNone(result)

    def test_role_remove_non_existent_group_project(self):
        # Simulate the user not being in keystone, the client the gracefully
        # handle this exception and send the request to remove the role since
        # keystone will validate.
        self.identity_sdk_client.find_group.side_effect = [
            sdk_exc.ResourceNotFound,
        ]
        arglist = [
            '--group',
            self.group.id,
            '--project',
            self.project.name,
            self.role.name,
        ]
        if self._is_inheritance_testcase():
            arglist.append('--inherited')
        verifylist = [
            ('user', None),
            ('group', self.group.id),
            ('system', None),
            ('domain', None),
            ('project', self.project.name),
            ('role', self.role.name),
            ('inherited', self._is_inheritance_testcase()),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'group': self.group.id,
            'project': self.project.id,
            'role': self.role.id,
            'inherited': self._is_inheritance_testcase(),
        }
        self.identity_sdk_client.unassign_project_role_from_group.assert_called_with(
            **kwargs
        )
        self.assertIsNone(result)

    def test_role_remove_domain_role_on_group_domain(self):
        self.role_with_domain = sdk_fakes.generate_fake_resource(
            resource_type=_role.Role,
            domain_id=self.domain.id,
            description=None,
        )
        self.identity_sdk_client.find_role.return_value = self.role_with_domain
        arglist = [
            '--group',
            self.group.name,
            '--domain',
            self.domain.name,
            self.role_with_domain.name,
        ]
        if self._is_inheritance_testcase():
            arglist.append('--inherited')
        verifylist = [
            ('user', None),
            ('group', self.group.name),
            ('domain', self.domain.name),
            ('project', None),
            ('role', self.role_with_domain.name),
            ('inherited', self._is_inheritance_testcase()),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'group': self.group.id,
            'domain': self.domain.id,
            'role': self.role_with_domain.id,
            'inherited': self._is_inheritance_testcase(),
        }
        self.identity_sdk_client.unassign_domain_role_from_group.assert_called_with(
            **kwargs
        )
        self.assertIsNone(result)

    def test_role_remove_with_error(self):
        arglist = [
            self.role.name,
        ]
        verifylist = [
            ('user', None),
            ('group', None),
            ('domain', None),
            ('project', None),
            ('role', self.role.name),
            ('inherited', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestRoleSet(identity_fakes.TestIdentityv3):
    def setUp(self):
        super().setUp()

        self.domain = sdk_fakes.generate_fake_resource(_domain.Domain)
        self.role_with_domain = sdk_fakes.generate_fake_resource(
            resource_type=_role.Role,
            domain_id=self.domain.id,
            description=None,
        )

        # Get the command object to test
        self.cmd = role.SetRole(self.app, None)

    def test_role_set_no_options(self):
        self.role = sdk_fakes.generate_fake_resource(
            resource_type=_role.Role,
            domain_id=None,
            description=None,
        )
        self.identity_sdk_client.find_role.return_value = self.role
        self.identity_sdk_client.update_role.return_value = self.role

        arglist = [
            '--name',
            'over',
            self.role.name,
        ]
        verifylist = [
            ('name', 'over'),
            ('role', self.role.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': 'over',
            'role': self.role.id,
        }
        self.identity_sdk_client.update_role.assert_called_with(**kwargs)
        self.assertIsNone(result)

    def test_role_set_domain_role(self):
        self.domain2 = sdk_fakes.generate_fake_resource(_domain.Domain)
        self.identity_sdk_client.find_domain.return_value = self.domain2

        self.identity_sdk_client.find_role.return_value = self.role_with_domain
        self.identity_sdk_client.update_role.return_value = (
            self.role_with_domain
        )

        arglist = [
            '--name',
            'over',
            '--domain',
            self.domain2.name,
            self.role_with_domain.name,
        ]
        verifylist = [
            ('name', 'over'),
            ('domain', self.domain2.name),
            ('role', self.role_with_domain.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': 'over',
            'role': self.role_with_domain.id,
            'domain_id': self.domain2.id,
        }
        self.identity_sdk_client.update_role.assert_called_with(**kwargs)
        self.assertIsNone(result)

    def test_role_set_description(self):
        self.identity_sdk_client.find_role.return_value = self.role_with_domain

        arglist = [
            '--name',
            'over',
            '--description',
            'role description',
            self.role_with_domain.name,
        ]
        verifylist = [
            ('name', 'over'),
            ('description', 'role description'),
            ('role', self.role_with_domain.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': 'over',
            'description': 'role description',
            'role': self.role_with_domain.id,
        }
        self.identity_sdk_client.update_role.assert_called_with(**kwargs)
        self.assertIsNone(result)

    def test_role_set_with_immutable(self):
        self.identity_sdk_client.find_role.return_value = self.role_with_domain

        arglist = [
            '--name',
            'over',
            '--immutable',
            self.role_with_domain.name,
        ]
        verifylist = [
            ('name', 'over'),
            ('immutable', True),
            ('role', self.role_with_domain.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': 'over',
            'role': self.role_with_domain.id,
            'options': {'immutable': True},
        }
        self.identity_sdk_client.update_role.assert_called_with(**kwargs)
        self.assertIsNone(result)

    def test_role_set_with_no_immutable(self):
        self.identity_sdk_client.find_role.return_value = self.role_with_domain

        arglist = [
            '--name',
            'over',
            '--no-immutable',
            self.role_with_domain.name,
        ]
        verifylist = [
            ('name', 'over'),
            ('immutable', False),
            ('role', self.role_with_domain.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': 'over',
            'role': self.role_with_domain.id,
            'options': {'immutable': False},
        }
        self.identity_sdk_client.update_role.assert_called_with(**kwargs)
        self.assertIsNone(result)


class TestRoleShow(identity_fakes.TestIdentityv3):
    domain = sdk_fakes.generate_fake_resource(_domain.Domain)

    def setUp(self):
        super().setUp()

        self.identity_sdk_client.find_domain.return_value = self.domain

        # Get the command object to test
        self.cmd = role.ShowRole(self.app, None)

    def test_role_show(self):
        self.role = sdk_fakes.generate_fake_resource(
            resource_type=_role.Role,
            domain_id=None,
            description=None,
        )
        self.identity_sdk_client.find_role.return_value = self.role

        arglist = [
            self.role.name,
        ]
        verifylist = [
            ('role', self.role.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.find_role.assert_called_with(
            name_or_id=self.role.name,
            domain_id=None,
            ignore_missing=False,
        )

        collist = ('id', 'name', 'domain_id', 'description')
        self.assertEqual(collist, columns)
        datalist = (
            self.role.id,
            self.role.name,
            None,
            None,
        )
        self.assertEqual(datalist, data)

    def test_role_show_domain_role(self):
        self.role_with_domain = sdk_fakes.generate_fake_resource(
            resource_type=_role.Role,
            domain_id=self.domain.id,
            description=None,
        )
        self.identity_sdk_client.find_role.return_value = self.role_with_domain

        arglist = [
            '--domain',
            self.domain.name,
            self.role_with_domain.id,
        ]
        verifylist = [
            ('domain', self.domain.name),
            ('role', self.role_with_domain.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.find_role.assert_called_with(
            name_or_id=self.role_with_domain.id,
            domain_id=self.domain.id,
            ignore_missing=False,
        )

        collist = ('id', 'name', 'domain_id', 'description')
        self.assertEqual(collist, columns)
        datalist = (
            self.role_with_domain.id,
            self.role_with_domain.name,
            self.domain.id,
            None,
        )
        self.assertEqual(datalist, data)



#   Copyright 2014 CERN.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import copy

from openstackclient.identity.v3 import federation_protocol
from openstackclient.tests.unit import fakes
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes


class TestProtocol(identity_fakes.TestFederatedIdentity):
    def setUp(self):
        super().setUp()

        federation_lib = self.identity_client.federation
        self.protocols_mock = federation_lib.protocols
        self.protocols_mock.reset_mock()


class TestProtocolCreate(TestProtocol):
    def setUp(self):
        super().setUp()

        proto = copy.deepcopy(identity_fakes.PROTOCOL_OUTPUT)
        resource = fakes.FakeResource(None, proto, loaded=True)
        self.protocols_mock.create.return_value = resource
        self.cmd = federation_protocol.CreateProtocol(self.app, None)

    def test_create_protocol(self):
        argslist = [
            identity_fakes.protocol_id,
            '--identity-provider',
            identity_fakes.idp_id,
            '--mapping',
            identity_fakes.mapping_id,
        ]

        verifylist = [
            ('federation_protocol', identity_fakes.protocol_id),
            ('identity_provider', identity_fakes.idp_id),
            ('mapping', identity_fakes.mapping_id),
        ]
        parsed_args = self.check_parser(self.cmd, argslist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.protocols_mock.create.assert_called_with(
            protocol_id=identity_fakes.protocol_id,
            identity_provider=identity_fakes.idp_id,
            mapping=identity_fakes.mapping_id,
        )

        collist = ('id', 'identity_provider', 'mapping')
        self.assertEqual(collist, columns)

        datalist = (
            identity_fakes.protocol_id,
            identity_fakes.idp_id,
            identity_fakes.mapping_id,
        )
        self.assertEqual(datalist, data)


class TestProtocolDelete(TestProtocol):
    def setUp(self):
        super().setUp()

        # This is the return value for utils.find_resource()
        self.protocols_mock.get.return_value = fakes.FakeResource(
            None,
            copy.deepcopy(identity_fakes.PROTOCOL_OUTPUT),
            loaded=True,
        )

        self.protocols_mock.delete.return_value = None
        self.cmd = federation_protocol.DeleteProtocol(self.app, None)

    def test_delete_identity_provider(self):
        arglist = [
            '--identity-provider',
            identity_fakes.idp_id,
            identity_fakes.protocol_id,
        ]
        verifylist = [
            ('federation_protocol', [identity_fakes.protocol_id]),
            ('identity_provider', identity_fakes.idp_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.protocols_mock.delete.assert_called_with(
            identity_fakes.idp_id, identity_fakes.protocol_id
        )
        self.assertIsNone(result)


class TestProtocolList(TestProtocol):
    def setUp(self):
        super().setUp()

        self.protocols_mock.get.return_value = fakes.FakeResource(
            None, identity_fakes.PROTOCOL_ID_MAPPING, loaded=True
        )

        self.protocols_mock.list.return_value = [
            fakes.FakeResource(
                None, identity_fakes.PROTOCOL_ID_MAPPING, loaded=True
            )
        ]

        self.cmd = federation_protocol.ListProtocols(self.app, None)

    def test_list_protocols(self):
        arglist = ['--identity-provider', identity_fakes.idp_id]
        verifylist = [('identity_provider', identity_fakes.idp_id)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.protocols_mock.list.assert_called_with(identity_fakes.idp_id)


class TestProtocolSet(TestProtocol):
    def setUp(self):
        super().setUp()
        self.protocols_mock.get.return_value = fakes.FakeResource(
            None, identity_fakes.PROTOCOL_OUTPUT, loaded=True
        )
        self.protocols_mock.update.return_value = fakes.FakeResource(
            None, identity_fakes.PROTOCOL_OUTPUT_UPDATED, loaded=True
        )

        self.cmd = federation_protocol.SetProtocol(self.app, None)

    def test_set_new_mapping(self):
        arglist = [
            identity_fakes.protocol_id,
            '--identity-provider',
            identity_fakes.idp_id,
            '--mapping',
            identity_fakes.mapping_id,
        ]
        verifylist = [
            ('identity_provider', identity_fakes.idp_id),
            ('federation_protocol', identity_fakes.protocol_id),
            ('mapping', identity_fakes.mapping_id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.protocols_mock.update.assert_called_with(
            identity_fakes.idp_id,
            identity_fakes.protocol_id,
            identity_fakes.mapping_id,
        )

        collist = ('id', 'identity_provider', 'mapping')
        self.assertEqual(collist, columns)

        datalist = (
            identity_fakes.protocol_id,
            identity_fakes.idp_id,
            identity_fakes.mapping_id_updated,
        )
        self.assertEqual(datalist, data)


class TestProtocolShow(TestProtocol):
    def setUp(self):
        super().setUp()
        self.protocols_mock.get.return_value = fakes.FakeResource(
            None, identity_fakes.PROTOCOL_OUTPUT, loaded=False
        )

        self.cmd = federation_protocol.ShowProtocol(self.app, None)

    def test_show_protocol(self):
        arglist = [
            identity_fakes.protocol_id,
            '--identity-provider',
            identity_fakes.idp_id,
        ]
        verifylist = [
            ('federation_protocol', identity_fakes.protocol_id),
            ('identity_provider', identity_fakes.idp_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.protocols_mock.get.assert_called_with(
            identity_fakes.idp_id, identity_fakes.protocol_id
        )

        collist = ('id', 'identity_provider', 'mapping')
        self.assertEqual(collist, columns)

        datalist = (
            identity_fakes.protocol_id,
            identity_fakes.idp_id,
            identity_fakes.mapping_id,
        )
        self.assertEqual(datalist, data)



# Copyright 2023 Red Hat.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstackclient.image.v2 import info
from openstackclient.tests.unit.image.v2 import fakes as image_fakes


class TestImportInfo(image_fakes.TestImagev2):
    import_info = image_fakes.create_one_import_info()

    def setUp(self):
        super().setUp()

        self.image_client.get_import_info.return_value = self.import_info

        self.cmd = info.ImportInfo(self.app, None)

    def test_import_info(self):
        arglist = []
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)

        self.image_client.get_import_info.assert_called()



#   Copyright 2015 Mirantis, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from unittest import mock

from openstack.compute.v2 import service as _service
from openstack.test import fakes as sdk_fakes
from osc_lib import exceptions

from openstackclient.compute.v2 import service
from openstackclient.tests.unit.compute.v2 import fakes as compute_fakes


class TestServiceDelete(compute_fakes.TestComputev2):
    def setUp(self):
        super().setUp()

        self.services = list(
            sdk_fakes.generate_fake_resources(_service.Service, count=2)
        )

        self.compute_client.delete_service.return_value = None

        # Get the command object to test
        self.cmd = service.DeleteService(self.app, None)

    def test_service_delete(self):
        arglist = [
            self.services[0].binary,
        ]
        verifylist = [
            ('service', [self.services[0].binary]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.compute_client.delete_service.assert_called_with(
            self.services[0].binary, ignore_missing=False
        )
        self.assertIsNone(result)

    def test_multi_services_delete(self):
        arglist = []
        for s in self.services:
            arglist.append(s.binary)
        verifylist = [
            ('service', arglist),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        calls = []
        for s in self.services:
            calls.append(mock.call(s.binary, ignore_missing=False))
        self.compute_client.delete_service.assert_has_calls(calls)
        self.assertIsNone(result)

    def test_multi_services_delete_with_exception(self):
        arglist = [
            self.services[0].binary,
            'unexist_service',
        ]
        verifylist = [('service', arglist)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        delete_mock_result = [None, exceptions.CommandError]
        self.compute_client.delete_service.side_effect = delete_mock_result

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual(
                '1 of 2 compute services failed to delete.', str(e)
            )

        self.compute_client.delete_service.assert_any_call(
            self.services[0].binary, ignore_missing=False
        )
        self.compute_client.delete_service.assert_any_call(
            'unexist_service', ignore_missing=False
        )


class TestServiceList(compute_fakes.TestComputev2):
    def setUp(self):
        super().setUp()

        self.service = sdk_fakes.generate_fake_resource(_service.Service)

        self.compute_client.services.return_value = [self.service]

        # Get the command object to test
        self.cmd = service.ListService(self.app, None)

    def test_service_list(self):
        arglist = [
            '--host',
            self.service.host,
            '--service',
            self.service.binary,
        ]
        verifylist = [
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.compute_client.services.assert_called_with(
            host=self.service.host,
            binary=self.service.binary,
        )

        expected_columns = (
            'ID',
            'Binary',
            'Host',
            'Zone',
            'Status',
            'State',
            'Updated At',
        )
        expected_data = [
            (
                self.service.id,
                self.service.binary,
                self.service.host,
                self.service.availability_zone,
                self.service.status,
                self.service.state,
                self.service.updated_at,
            )
        ]

        self.assertEqual(expected_columns, columns)
        self.assertEqual(expected_data, list(data))

    def test_service_list_with_long_option(self):
        arglist = [
            '--host',
            self.service.host,
            '--service',
            self.service.binary,
            '--long',
        ]
        verifylist = [
            ('host', self.service.host),
            ('service', self.service.binary),
            ('long', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.compute_client.services.assert_called_with(
            host=self.service.host,
            binary=self.service.binary,
        )

        expected_columns = (
            'ID',
            'Binary',
            'Host',
            'Zone',
            'Status',
            'State',
            'Updated At',
            'Disabled Reason',
        )
        expected_data = [
            (
                self.service.id,
                self.service.binary,
                self.service.host,
                self.service.availability_zone,
                self.service.status,
                self.service.state,
                self.service.updated_at,
                self.service.disabled_reason,
            )
        ]

        self.assertEqual(expected_columns, columns)
        self.assertEqual(expected_data, list(data))

    def test_service_list_with_long_option_2_11(self):
        self.set_compute_api_version('2.11')

        arglist = [
            '--host',
            self.service.host,
            '--service',
            self.service.binary,
            '--long',
        ]
        verifylist = [
            ('host', self.service.host),
            ('service', self.service.binary),
            ('long', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.compute_client.services.assert_called_with(
            host=self.service.host,
            binary=self.service.binary,
        )

        # In 2.11 there is also a forced_down column.
        expected_columns = (
            'ID',
            'Binary',
            'Host',
            'Zone',
            'Status',
            'State',
            'Updated At',
            'Disabled Reason',
            'Forced Down',
        )
        expected_data = [
            (
                self.service.id,
                self.service.binary,
                self.service.host,
                self.service.availability_zone,
                self.service.status,
                self.service.state,
                self.service.updated_at,
                self.service.disabled_reason,
                self.service.is_forced_down,
            )
        ]

        self.assertEqual(expected_columns, columns)
        self.assertEqual(expected_data, list(data))


class TestServiceSet(compute_fakes.TestComputev2):
    def setUp(self):
        super().setUp()

        self.service = sdk_fakes.generate_fake_resource(_service.Service)

        self.compute_client.enable_service.return_value = self.service
        self.compute_client.disable_service.return_value = self.service

        self.cmd = service.SetService(self.app, None)

    def test_set_nothing(self):
        arglist = [
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.compute_client.enable_service.assert_not_called()
        self.compute_client.disable_service.assert_not_called()
        self.assertIsNone(result)

    def test_service_set_enable(self):
        arglist = [
            '--enable',
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('enable', True),
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.compute_client.enable_service.assert_called_with(
            None, self.service.host, self.service.binary
        )
        self.assertIsNone(result)

    def test_service_set_disable(self):
        arglist = [
            '--disable',
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('disable', True),
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.compute_client.disable_service.assert_called_with(
            None, self.service.host, self.service.binary, None
        )
        self.assertIsNone(result)

    def test_service_set_disable_with_reason(self):
        reason = 'earthquake'
        arglist = [
            '--disable',
            '--disable-reason',
            reason,
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('disable', True),
            ('disable_reason', reason),
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.compute_client.disable_service.assert_called_with(
            None, self.service.host, self.service.binary, reason
        )
        self.assertIsNone(result)

    def test_service_set_only_with_disable_reason(self):
        reason = 'earthquake'
        arglist = [
            '--disable-reason',
            reason,
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('disable_reason', reason),
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        try:
            self.cmd.take_action(parsed_args)
            self.fail("CommandError should be raised.")
        except exceptions.CommandError as e:
            self.assertEqual(
                "Cannot specify option --disable-reason without "
                "--disable specified.",
                str(e),
            )

    def test_service_set_enable_with_disable_reason(self):
        reason = 'earthquake'
        arglist = [
            '--enable',
            '--disable-reason',
            reason,
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('enable', True),
            ('disable_reason', reason),
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        try:
            self.cmd.take_action(parsed_args)
            self.fail("CommandError should be raised.")
        except exceptions.CommandError as e:
            self.assertEqual(
                "Cannot specify option --disable-reason without "
                "--disable specified.",
                str(e),
            )

    def test_service_set_state_up(self):
        self.set_compute_api_version('2.11')

        arglist = [
            '--up',
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('up', True),
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.compute_client.update_service_forced_down.assert_called_once_with(
            None, self.service.host, self.service.binary, False
        )
        self.assertNotCalled(self.compute_client.enable_service)
        self.assertNotCalled(self.compute_client.disable_service)
        self.assertIsNone(result)

    def test_service_set_state_down(self):
        self.set_compute_api_version('2.11')

        arglist = [
            '--down',
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('down', True),
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.compute_client.update_service_forced_down.assert_called_once_with(
            None, self.service.host, self.service.binary, True
        )
        self.assertNotCalled(self.compute_client.enable_service)
        self.assertNotCalled(self.compute_client.disable_service)
        self.assertIsNone(result)

    def test_service_set_enable_and_state_down(self):
        self.set_compute_api_version('2.11')

        arglist = [
            '--enable',
            '--down',
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('enable', True),
            ('down', True),
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.compute_client.enable_service.assert_called_once_with(
            None, self.service.host, self.service.binary
        )
        self.compute_client.update_service_forced_down.assert_called_once_with(
            None, self.service.host, self.service.binary, True
        )
        self.assertIsNone(result)

    def test_service_set_enable_and_state_down_with_exception(self):
        self.set_compute_api_version('2.11')

        arglist = [
            '--enable',
            '--down',
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('enable', True),
            ('down', True),
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        with mock.patch.object(
            self.compute_client, 'enable_service', side_effect=Exception()
        ):
            self.assertRaises(
                exceptions.CommandError, self.cmd.take_action, parsed_args
            )
            self.compute_client.update_service_forced_down.assert_called_once_with(
                None, self.service.host, self.service.binary, True
            )

    def test_service_set_disable_down(self):
        # Tests disabling and forcing down a compute service with microversion
        # 2.53 which requires looking up the service by host and binary.
        self.set_compute_api_version('2.53')

        arglist = [
            '--disable',
            '--down',
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('disable', True),
            ('down', True),
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        service_id = '339478d0-0b95-4a94-be63-d5be05dfeb1c'
        self.compute_client.services.return_value = [mock.Mock(id=service_id)]
        result = self.cmd.take_action(parsed_args)
        self.compute_client.disable_service.assert_called_once_with(
            service_id, self.service.host, self.service.binary, None
        )
        self.compute_client.update_service_forced_down.assert_called_once_with(
            service_id, self.service.host, self.service.binary, True
        )
        self.assertIsNone(result)

    def test_service_set_disable_reason(self):
        # Tests disabling with reason a compute service with microversion
        # 2.53 which requires looking up the service by host and binary.
        self.set_compute_api_version('2.53')

        reason = 'earthquake'
        arglist = [
            '--disable',
            '--disable-reason',
            reason,
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('disable', True),
            ('disable_reason', reason),
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        service_id = '339478d0-0b95-4a94-be63-d5be05dfeb1c'
        self.compute_client.services.return_value = [mock.Mock(id=service_id)]
        result = self.cmd.take_action(parsed_args)
        self.compute_client.disable_service.assert_called_once_with(
            service_id, self.service.host, self.service.binary, reason
        )
        self.assertIsNone(result)

    def test_service_set_enable_up(self):
        # Tests enabling and bringing up a compute service with microversion
        # 2.53 which requires looking up the service by host and binary.
        self.set_compute_api_version('2.53')

        arglist = [
            '--enable',
            '--up',
            self.service.host,
            self.service.binary,
        ]
        verifylist = [
            ('enable', True),
            ('up', True),
            ('host', self.service.host),
            ('service', self.service.binary),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        service_id = '339478d0-0b95-4a94-be63-d5be05dfeb1c'
        self.compute_client.services.return_value = [mock.Mock(id=service_id)]
        result = self.cmd.take_action(parsed_args)
        self.compute_client.enable_service.assert_called_once_with(
            service_id, self.service.host, self.service.binary
        )
        self.compute_client.update_service_forced_down.assert_called_once_with(
            service_id, self.service.host, self.service.binary, False
        )
        self.assertIsNone(result)

    def test_service_set_find_service_by_host_and_binary_no_results(self):
        # Tests that no compute services are found by host and binary.
        self.compute_client.services.return_value = []
        ex = self.assertRaises(
            exceptions.CommandError,
            self.cmd._find_service_by_host_and_binary,
            self.compute_client,
            'fake-host',
            'nova-compute',
        )
        self.assertIn(
            'Compute service for host "fake-host" and binary '
            '"nova-compute" not found.',
            str(ex),
        )

    def test_service_set_find_service_by_host_and_binary_many_results(self):
        # Tests that more than one compute service is found by host and binary.
        self.compute_client.services.return_value = [
            mock.Mock(),
            mock.Mock(),
        ]
        ex = self.assertRaises(
            exceptions.CommandError,
            self.cmd._find_service_by_host_and_binary,
            self.compute_client,
            'fake-host',
            'nova-compute',
        )
        self.assertIn(
            'Multiple compute services found for host "fake-host" '
            'and binary "nova-compute". Unable to proceed.',
            str(ex),
        )



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Service action implementations"""

import logging

from openstack import utils as sdk_utils
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _


LOG = logging.getLogger(__name__)


class DeleteService(command.Command):
    _description = _("Delete compute service(s)")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "service",
            metavar="<service>",
            nargs='+',
            help=_(
                "Compute service(s) to delete (ID only). If using "
                "``--os-compute-api-version`` 2.53 or greater, the ID is "
                "a UUID which can be retrieved by listing compute services "
                "using the same 2.53+ microversion. "
                "If deleting a compute service, be sure to stop the actual "
                "compute process on the physical host before deleting the "
                "service with this command. Failing to do so can lead to "
                "the running service re-creating orphaned compute_nodes "
                "table records in the database."
            ),
        )
        return parser

    def take_action(self, parsed_args):
        compute_client = self.app.client_manager.compute
        result = 0
        for s in parsed_args.service:
            try:
                compute_client.delete_service(s, ignore_missing=False)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete compute service with "
                        "ID '%(service)s': %(e)s"
                    ),
                    {'service': s, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.service)
            msg = _(
                "%(result)s of %(total)s compute services failed to delete."
            ) % {'result': result, 'total': total}
            raise exceptions.CommandError(msg)


class ListService(command.Lister):
    _description = _(
        """List compute services.

Using ``--os-compute-api-version`` 2.53 or greater will return the ID as a UUID
value which can be used to uniquely identify the service in a multi-cell
deployment."""
    )

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "--host",
            metavar="<host>",
            help=_("List services on specified host (name only)"),
        )
        parser.add_argument(
            "--service",
            metavar="<service>",
            help=_(
                "List only specified service binaries (name only). For "
                "example, ``nova-compute``, ``nova-conductor``, etc."
            ),
        )
        parser.add_argument(
            "--long",
            action="store_true",
            default=False,
            help=_("List additional fields in output"),
        )
        return parser

    def take_action(self, parsed_args):
        compute_client = self.app.client_manager.compute
        columns: tuple[str, ...] = (
            "id",
            "binary",
            "host",
            "availability_zone",
            "status",
            "state",
            "updated_at",
        )
        column_headers: tuple[str, ...] = (
            "ID",
            "Binary",
            "Host",
            "Zone",
            "Status",
            "State",
            "Updated At",
        )
        if parsed_args.long:
            columns += ("disabled_reason",)
            column_headers += ("Disabled Reason",)
            if sdk_utils.supports_microversion(compute_client, '2.11'):
                columns += ("is_forced_down",)
                column_headers += ("Forced Down",)

        data = compute_client.services(
            host=parsed_args.host, binary=parsed_args.service
        )
        return (
            column_headers,
            (utils.get_item_properties(s, columns) for s in data),
        )


class SetService(command.Command):
    _description = _("Set compute service properties")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument("host", metavar="<host>", help=_("Name of host"))
        parser.add_argument(
            "service",
            metavar="<service>",
            help=_(
                "Name of service (Binary name), for example ``nova-compute``"
            ),
        )
        enabled_group = parser.add_mutually_exclusive_group()
        enabled_group.add_argument(
            "--enable", action="store_true", help=_("Enable service")
        )
        enabled_group.add_argument(
            "--disable", action="store_true", help=_("Disable service")
        )
        parser.add_argument(
            "--disable-reason",
            default=None,
            metavar="<reason>",
            help=_(
                "Reason for disabling the service (in quotes). "
                "Should be used with --disable option."
            ),
        )
        up_down_group = parser.add_mutually_exclusive_group()
        up_down_group.add_argument(
            '--up',
            action='store_true',
            help=_(
                'Force up service. Requires ``--os-compute-api-version`` '
                '2.11 or greater.'
            ),
        )
        up_down_group.add_argument(
            '--down',
            action='store_true',
            help=_(
                'Force down service. Requires ``--os-compute-api-version`` '
                '2.11 or greater.'
            ),
        )
        return parser

    @staticmethod
    def _find_service_by_host_and_binary(compute_client, host, binary):
        """Utility method to find a compute service by host and binary

        :param host: the name of the compute service host
        :param binary: the compute service binary, e.g. nova-compute
        :returns: The service.
        :raises: CommandError if no or multiple results were found
        """
        services = list(compute_client.services(host=host, binary=binary))
        # Did we find anything?
        if not len(services):
            msg = _(
                'Compute service for host "%(host)s" and binary '
                '"%(binary)s" not found.'
            ) % {'host': host, 'binary': binary}
            raise exceptions.CommandError(msg)
        # Did we find more than one result? This should not happen but let's
        # be safe.
        if len(services) > 1:
            # TODO(mriedem): If we have an --id option for 2.53+ then we can
            # say to use that option to uniquely identify the service.
            msg = _(
                'Multiple compute services found for host "%(host)s" and '
                'binary "%(binary)s". Unable to proceed.'
            ) % {'host': host, 'binary': binary}
            raise exceptions.CommandError(msg)
        return services[0]

    def take_action(self, parsed_args):
        compute_client = self.app.client_manager.compute

        if (
            parsed_args.enable or not parsed_args.disable
        ) and parsed_args.disable_reason:
            msg = _(
                "Cannot specify option --disable-reason without "
                "--disable specified."
            )
            raise exceptions.CommandError(msg)

        # Starting with microversion 2.53, there is a single
        # PUT /os-services/{service_id} API for updating nova-compute
        # services. If 2.53+ is used we need to find the nova-compute
        # service using the --host and --service (binary) values.
        requires_service_id = sdk_utils.supports_microversion(
            compute_client, '2.53'
        )
        service_id = None
        if requires_service_id:
            # TODO(mriedem): Add an --id option so users can pass the service
            # id (as a uuid) directly rather than make us look it up using
            # host/binary.
            service_id = SetService._find_service_by_host_and_binary(
                compute_client, parsed_args.host, parsed_args.service
            ).id

        result = 0
        enabled = None
        try:
            if parsed_args.enable:
                enabled = True
            if parsed_args.disable:
                enabled = False

            if enabled is not None:
                if enabled:
                    compute_client.enable_service(
                        service_id, parsed_args.host, parsed_args.service
                    )
                else:
                    compute_client.disable_service(
                        service_id,
                        parsed_args.host,
                        parsed_args.service,
                        parsed_args.disable_reason,
                    )
        except Exception:
            status = "enabled" if enabled else "disabled"
            LOG.error("Failed to set service status to %s", status)
            result += 1

        force_down = None
        if parsed_args.down:
            force_down = True
        if parsed_args.up:
            force_down = False
        if force_down is not None:
            if not sdk_utils.supports_microversion(compute_client, '2.11'):
                msg = _('--os-compute-api-version 2.11 or later is required')
                raise exceptions.CommandError(msg)
            try:
                compute_client.update_service_forced_down(
                    service_id,
                    parsed_args.host,
                    parsed_args.service,
                    force_down,
                )
            except Exception:
                state = "down" if force_down else "up"
                LOG.error("Failed to set service state to %s", state)
                result += 1

        if result > 0:
            msg = _(
                "Compute service %(service)s of host %(host)s failed to set."
            ) % {"service": parsed_args.service, "host": parsed_args.host}
            raise exceptions.CommandError(msg)



#   Copyright 2017 Huawei, Inc. All rights reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from openstackclient.tests.functional.compute.v2 import common


class ServerEventTests(common.ComputeTestCase):
    """Functional tests for server event"""

    def setUp(self):
        super().setUp()

        # NOTE(dtroyer): As long as these tests are read-only we can get away
        #                with using the same server instance for all of them.
        cmd_output = self.server_create()
        self.server_id = cmd_output.get('id')
        self.server_name = cmd_output['name']

    def test_server_event_list_and_show(self):
        """Test list, show server event"""
        # Test 'server event list' for creating
        cmd_output = self.openstack(
            'server event list ' + self.server_name,
            parse_output=True,
        )
        request_id = None
        for each_event in cmd_output:
            self.assertNotIn('Message', each_event)
            self.assertNotIn('Project ID', each_event)
            self.assertNotIn('User ID', each_event)
            if each_event.get('Action') == 'create':
                self.assertEqual(self.server_id, each_event.get('Server ID'))
                request_id = each_event.get('Request ID')
                break
        self.assertIsNotNone(request_id)
        # Test 'server event show' for creating
        cmd_output = self.openstack(
            'server event show ' + self.server_name + ' ' + request_id,
            parse_output=True,
        )
        self.assertEqual(request_id, cmd_output.get('request_id'))
        self.assertEqual('create', cmd_output.get('action'))
        self.assertIsNotNone(cmd_output.get('events'))
        self.assertIsInstance(cmd_output.get('events'), list)

        # Reboot server, trigger reboot event
        self.openstack('server reboot --wait ' + self.server_name)
        # Test 'server event list --long' for rebooting
        cmd_output = self.openstack(
            'server event list --long ' + self.server_name,
            parse_output=True,
        )
        request_id = None
        for each_event in cmd_output:
            self.assertIn('Message', each_event)
            self.assertIn('Project ID', each_event)
            self.assertIn('User ID', each_event)
            if each_event.get('Action') == 'reboot':
                request_id = each_event.get('Request ID')
                self.assertEqual(self.server_id, each_event.get('Server ID'))
                break
        self.assertIsNotNone(request_id)
        # Test 'server event show' for rebooting
        cmd_output = self.openstack(
            'server event show ' + self.server_name + ' ' + request_id,
            parse_output=True,
        )
        self.assertEqual(request_id, cmd_output.get('request_id'))
        self.assertEqual('reboot', cmd_output.get('action'))
        self.assertIsNotNone(cmd_output.get('events'))
        self.assertIsInstance(cmd_output.get('events'), list)

    def test_server_event_list_and_show_deleted_server(self):
        # Need to create a new server that will not attempt cleanup because we
        # are going to delete it during the test and cleanup would fail with
        # 404.
        cmd_output = self.server_create(cleanup=False)
        server_id = cmd_output['id']
        # Delete the server
        self.openstack('server delete --wait ' + server_id)
        # And verify we can get the event list after it's deleted
        # Test 'server event list' for deleting
        cmd_output = self.openstack(
            '--os-compute-api-version 2.21 server event list ' + server_id,
            parse_output=True,
        )
        request_id = None
        for each_event in cmd_output:
            self.assertNotIn('Message', each_event)
            self.assertNotIn('Project ID', each_event)
            self.assertNotIn('User ID', each_event)
            if each_event.get('Action') == 'delete':
                self.assertEqual(server_id, each_event.get('Server ID'))
                request_id = each_event.get('Request ID')
                break
        self.assertIsNotNone(request_id)
        # Test 'server event show' for deleting
        cmd_output = self.openstack(
            '--os-compute-api-version 2.21 '
            'server event show ' + server_id + ' ' + request_id,
            parse_output=True,
        )
        self.assertEqual(request_id, cmd_output.get('request_id'))
        self.assertEqual('delete', cmd_output.get('action'))
        self.assertIsNotNone(cmd_output.get('events'))
        self.assertIsInstance(cmd_output.get('events'), list)



# Copyright 2023 Red Hat.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from openstackclient.tests.functional.image import base


class InfoTests(base.BaseImageTests):
    """Functional tests for Info commands"""

    def setUp(self):
        super().setUp()

    def tearDown(self):
        super().tearDown()

    def test_image_import_info(self):
        output = self.openstack('image import info', parse_output=True)
        self.assertIsNotNone(output['import-methods'])
